// Copyright 2019 Kina Wakasa
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'dart:async';
import 'dart:collection';
import 'dart:io';

import 'package:logging/logging.dart' show Level, Logger, LogRecord;
import 'package:uuid/uuid.dart' show Uuid;

import 'job.dart';
import 'pipeline.dart';
import 'service.dart';

enum RunnerStatus { Stopped, Stopping, Running, Closed }

class Runner {
  final String name;
  final _pipelines = Queue<Pipeline>();
  var _status = RunnerStatus.Stopped;
  StreamController<int> _controller;
  int _capacity;
  int _flow = 0;
  Service srv;
  bool isLog;
  IOSink _dumpFile;

  int get capacity => _capacity;

  set capacity(int newCapacity) {
    _capacity = newCapacity;
    update();
  }

  int get flow => _flow;

  RunnerStatus get status => _status;

  Runner(this.name, this.srv, {capacity = 1, this.isLog = false, dumpFile = ''})
      : _capacity = capacity {
    if (dumpFile.isNotEmpty) {
      _dumpFile = File(dumpFile).openWrite(mode: FileMode.append);
    }
    if (isLog && srv['logger'] == null) {
      srv['logger'] = Logger(name)
        ..onRecord.listen((LogRecord rec) {
          print('${rec.level.name}: ${rec.time}: ${rec.message}');
        });
    }
    if (srv['uuid'] == null) {
      srv['uuid'] = Uuid();
    }
  }

  void submit(Job job) {
    submitGroup(() sync* {
      yield job;
    }());
  }

  void submitGroup(Iterable<Job> group) {
    submitQueue(group.iterator);
  }

  void submitQueue(Iterator<Job> queue) {
    _pipelines.add(Pipeline(srv['uuid'].v1(), queue));
    update();
  }

  Future<bool> start() async {
    if (_status != RunnerStatus.Stopped) {
      return false;
    }
    _controller = StreamController<int>();
    _status = RunnerStatus.Running;
    update();
    await for (var _ in _controller.stream);
    return true;
  }

  bool pause() {
    if (_status != RunnerStatus.Running) {
      return false;
    }
    _status = RunnerStatus.Stopping;
    update();
    return true;
  }

  bool clear() {
    if (_status != RunnerStatus.Stopped) {
      return false;
    }
    _pipelines.clear();
    return true;
  }

  Future<bool> close() async {
    if (_status != RunnerStatus.Stopped) {
      return false;
    }
    _status = RunnerStatus.Closed;
    if (_dumpFile != null) {
      await _dumpFile.flush();
      await _dumpFile.close();
    }
    return true;
  }

  void _log(Level logLevel, var message) {
    if (isLog && srv['logger'] != null) {
      srv['logger']..log(logLevel, message);
    }
  }

  void update() {
    while (_status == RunnerStatus.Running &&
        _pipelines.isNotEmpty &&
        _flow < _capacity) {
      var pipeline = _pipelines.removeFirst();
      if (pipeline.moveNext()) {
        ++_flow;
        var job = pipeline.current;
        var desc =
            'Job #${pipeline.count} ${job.name} in Pipeline ${pipeline.uuid}';
        _log(Level.INFO, 'Running $desc.');
        job.exec(this, srv).then((var value) {
          --_flow;
          _log(Level.INFO, '$desc Done.');
          update();
        }).catchError((var error) {
          --_flow;
          _log(Level.SEVERE, '$desc throw error ${error}');
          if (_dumpFile != null && job.isDump) {
            _dumpFile.writeln(job.dump());
          }
          update();
        });
        _pipelines.add(pipeline);
      }
    }
    if (_flow == 0 &&
        _status != RunnerStatus.Stopped &&
        _status != RunnerStatus.Closed) {
      _status = RunnerStatus.Stopped;
      _controller.close();
    }
  }
}
